import struct
import json
import re

from quarry.data.chat_styles import code_by_name, code_by_prop
from quarry.utils import types
from quarry.utils.errors import ProtocolError

# Python 3 compat
try:
    basestring
except NameError:  # pragma: no cover
    basestring = str


class BufferUnderrun(Exception):
    pass


class Buffer(object):
    buff = b""
    pos = 0

    def __len__(self):
        return len(self.buff) - self.pos

    def add(self, data):
        """
        Add some bytes to the end of the buffer.
        """

        self.buff += data

    def save(self):
        """
        Saves the buffer contents.
        """

        self.buff = self.buff[self.pos:]
        self.pos = 0

    def restore(self):
        """
        Restores the buffer contents to its state when :meth:`save` was last
        called.
        """

        self.pos = 0

    def discard(self):
        """
        Discards the entire buffer contents.
        """

        self.pos = len(self.buff)

    def read(self, length=None):
        """
        Read *length* bytes from the beginning of the buffer buffer, or all
        bytes if *length* is ``None``
        """

        if length is None:
            data = self.buff[self.pos:]
            self.pos = len(self.buff)
        else:
            if self.pos + length > len(self.buff):
                raise BufferUnderrun()

            data = self.buff[self.pos:self.pos+length]
            self.pos += length

        return data

    def unpack(self, fmt):
        """
        Unpack a struct from the buffer. The format accepted is the same as
        for ``struct.unpack()``.
        """

        fmt = ">"+fmt
        length = struct.calcsize(fmt)
        fields = struct.unpack(fmt, self.read(length))
        if len(fields) == 1:
            fields = fields[0]
        return fields

    def unpack_string(self):
        """
        Unpack a Minecraft string (varint-prefixed utf8) from the buffer.
        """

        length = self.unpack_varint(max_bits=16)
        text = self.read(length).decode("utf-8")
        return text

    def unpack_json(self):
        """
        Unpack a Minecraft string from the buffer and interpret it as JSON.
        """

        obj = json.loads(self.unpack_string())
        return obj

    def unpack_chat(self, strip_styles=True):
        """
        Unpack a Minecraft chat message from the buffer. Minecraft uses a
        JSON format to send chat messages; this method retrieves a plaintext
        representation, optionally including styles encoded using oldschool
        chat codes (U+00A7 plus one character). For more control, use
        :meth:`unpack_json`.
        """

        def parse(obj):
            if isinstance(obj, basestring):
                return obj
            if isinstance(obj, list):
                return "".join((parse(e) for e in obj))
            if isinstance(obj, dict):
                text = ""
                for prop, code in code_by_prop.items():
                    if obj.get(prop):
                        text += u"\u00a7" + code
                if "color" in obj:
                    text += u"\u00a7" + code_by_name[obj["color"]]
                if "translate" in obj:
                    text += obj["translate"]
                    if "with" in obj:
                        args = ", ".join((parse(e) for e in obj["with"]))
                        text += "{%s}" % args
                if "text" in obj:
                    text += obj["text"]
                if "extra" in obj:
                    text += parse(obj["extra"])
                return text

        text = parse(self.unpack_json())
        if strip_styles:
            text = self.strip_chat_styles(text)
        return text

    def unpack_varint(self, max_bits=32, signed=False):
        """
        Unpacks a varint from the buffer.
        """

        number = 0
        for i in range(5):
            b = self.unpack("B")
            number |= (b & 0x7F) << 7*i
            if not b & 0x80:
                break

        if number & (1<<31):
            number -= 1<<32
            if not signed:
                raise ProtocolError("varint cannot be negative: %d" % number)

        number_min = -1 << (max_bits - 1)
        number_max = +1 << (max_bits - 1)
        if not (number_min <= number < number_max):
            raise ProtocolError("varint does not fit in range: %d <= %d < %d"
                                % (number_min, number, number_max))

        return number

    def unpack_uuid(self):
        """
        Unpacks a UUID from the buffer.
        """

        return types.UUID.from_bytes(self.read(16))

    def unpack_position(self):
        """
        Unpacks a Position from the buffer.
        """

        def unpack_twos_comp(bits, number):
            if (number & (1 << (bits - 1))) != 0:
                number = number - (1 << bits)
            return number

        number = self.unpack('Q')
        x = unpack_twos_comp(26, (number >> 38))
        y = unpack_twos_comp(12, (number >> 26 & 0xFFF))
        z = unpack_twos_comp(26, (number & 0x3FFFFFF))
        return x, y, z

    def unpack_slot(self):
        """
        Unpacks a Slot from the buffer.
        """

        slot = {}
        slot['id'] = self.unpack('h')
        if slot['id'] != -1:
            slot['count'] = self.unpack('b')
            slot['damage'] = self.unpack('h')
            slot['tag'] = self.unpack_nbt()

        return slot

    def unpack_nbt(self):
        """
        Unpacks NBT tag(s) from the buffer.
        """

        from quarry.utils import nbt
        return nbt.NamedTag.from_buff(self)

    @classmethod
    def pack(cls, fmt, *fields):
        """
        Pack *fields* into a struct. The format accepted is the same as for
        ``struct.pack()``.
        """

        return struct.pack(">"+fmt, *fields)

    @classmethod
    def pack_string(cls, text):
        """
        Pack a Minecraft string (varint-prefixed utf8).
        """

        text = text.encode("utf-8")
        return cls.pack_varint(len(text), max_bits=16) + text

    @classmethod
    def pack_json(cls, obj):
        """
        Serialize an object to JSON and pack it to a Minecraft string.
        """

        return cls.pack_string(json.dumps(obj))

    @classmethod
    def pack_chat(cls, text):
        """
        Pack a Minecraft chat message. This method accepts plaintext; to send
        colours and other formatting use :meth:`pack_json`.
        """

        return cls.pack_json({"text": text})

    @classmethod
    def pack_varint(cls, number, max_bits=32, signed=False):
        """
        Packs a varint.
        """

        number_min = -1 << (max_bits - 1)
        number_max = +1 << (max_bits - 1)
        if not (number_min <= number < number_max):
            raise ProtocolError("varint does not fit in range: %d <= %d < %d"
                                % (number_min, number, number_max))

        if number < 0:
            if signed:
                number += 1<<32
            else:
                raise ProtocolError("varint cannot be negative: %d" % number)

        out = b""
        for i in range(5):
            b = number & 0x7F
            number >>= 7
            out += cls.pack("B", b | (0x80 if number > 0 else 0))
            if number == 0:
                break
        return out

    @classmethod
    def pack_uuid(cls, uuid):
        """
        Packs a UUID.
        """

        return uuid.to_bytes()

    @classmethod
    def pack_nbt(cls, tag):
        return tag.to_bytes()

    @classmethod
    def strip_chat_styles(cls, text):
        """
        Strips chat styles from text.
        """
        return re.sub(u"\u00A7.", "", text)


import gzip

from .buffer import Buffer

_kinds = {}
_ids = {}


# Base types --------------------------------------------------------------------------------------

class _Tag(object):
    def __init__(self, value):
        self.value = value

    @classmethod
    def from_buff(cls, buff):
        raise NotImplementedError

    def to_bytes(self):
        raise NotImplementedError

    def to_obj(self):
        return self.value

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.value)


class _DataTag(_Tag):
    fmt = None

    @classmethod
    def from_buff(cls, buff):
        return cls(buff.unpack(cls.fmt))

    def to_bytes(self):
        return Buffer.pack(self.fmt, self.value)


class _ArrayTag(_Tag):
    inner_kind = None

    def __init__(self, value, inner_kind):
        super(_ArrayTag, self).__init__(value)
        self.inner_kind = inner_kind

    @classmethod
    def from_buff(cls, buff, inner_kind=None):
        if inner_kind is None:
            inner_kind = cls.inner_kind
        array_length = buff.unpack('i')
        return cls([inner_kind.from_buff(buff).value for _ in range(array_length)], inner_kind)

    def to_bytes(self):
        return (
            Buffer.pack('i', len(self.value)) +
            b"".join(
                self.inner_kind(elem).to_bytes() for elem in self.value))

    def to_obj(self):
        return [self.inner_kind(elem).to_obj() for elem in self.value]

    def __repr__(self):
        return "%s<%s>(%s)" % (type(self).__name__, self.inner_kind.__name__, self.value)


# Special types -----------------------------------------------------------------------------------

class NamedTag(_Tag):
    def __init__(self, value, name):
        super(NamedTag, self).__init__(value)
        self.name = name

    @classmethod
    def from_buff(cls, buff):
        kind_id = buff.unpack('b')
        if kind_id == 0:
            return None
        kind = _kinds[kind_id]
        name = TagString.from_buff(buff).value
        value = kind.from_buff(buff)
        return cls(value, name)

    def to_bytes(self):
        return (
            Buffer.pack('b', _ids[type(self.value)]) +
            TagString(self.name).to_bytes() +
            self.value.to_bytes())

    def to_obj(self):
        return self.name, self.value.to_obj()

    def __repr__(self):
        return "%s(%r, %r)" % (type(self).__name__, self.name, self.value)


# NBT tags ----------------------------------------------------------------------------------------

class TagByte(_DataTag):
    fmt = 'b'


class TagShort(_DataTag):
    fmt = 'h'


class TagInt(_DataTag):
    fmt = 'i'


class TagLong(_DataTag):
    fmt = 'q'


class TagFloat(_DataTag):
    fmt = 'f'


class TagDouble(_DataTag):
    fmt = 'd'


class TagString(_Tag):

    @classmethod
    def from_buff(cls, buff):
        string_length = buff.unpack('H')
        return cls(buff.read(string_length).decode('utf8'))

    def to_bytes(self):
        data = self.value.encode('utf8')
        return Buffer.pack('H', len(data)) + data


class TagByteArray(_ArrayTag):
    inner_kind = TagByte


class TagIntArray(_ArrayTag):
    inner_kind = TagInt


class TagList(_ArrayTag):

    @classmethod
    def from_buff(cls, buff):
        inner_kind_id = buff.unpack('b')
        inner_kind = _kinds[inner_kind_id]
        return super(TagList, cls).from_buff(buff, inner_kind)

    def to_bytes(self):
        return Buffer.pack('b', _ids[self.inner_kind]) + super(TagList, self).to_bytes()


class TagCompound(_Tag):

    @classmethod
    def from_buff(cls, buff):
        value = []
        while True:
            tag = NamedTag.from_buff(buff)
            if tag is None:
                return cls(value)
            value.append(tag)

    def to_bytes(self):
        return b"".join(tag.to_bytes() for tag in self.value) + Buffer.pack('b', 0)

    def to_obj(self):
        return dict(tag.to_obj() for tag in self.value)


# Register tags -----------------------------------------------------------------------------------

_kinds[0] = type(None)
_kinds[1] = TagByte
_kinds[2] = TagShort
_kinds[3] = TagInt
_kinds[4] = TagLong
_kinds[5] = TagFloat
_kinds[6] = TagDouble
_kinds[7] = TagByteArray
_kinds[8] = TagString
_kinds[9] = TagList
_kinds[10] = TagCompound
_kinds[11] = TagIntArray
_ids.update({v: k for k, v in _kinds.items()})

# Files -------------------------------------------------------------------------------------------

class NBTFile(object):
    def __init__(self, root_tag):
        self.root_tag = root_tag

    @classmethod
    def load(cls, path):
        with gzip.open(path, 'rb') as fd:
            buff = Buffer()
            buff.add(fd.read())
            return cls(NamedTag.from_buff(buff))

    def save(self, path):
        with gzip.open(path, 'wb') as fd:
            fd.write(self.root_tag.to_bytes())



import base64

from twisted.internet import reactor, defer

from quarry.net.protocol import Factory, Protocol, protocol_modes
from quarry.mojang import auth
from quarry.utils import crypto, types
from quarry.utils.errors import ProtocolError


class ServerProtocol(Protocol):
    """This class represents a connection with a client"""

    recv_direction = "upstream"
    send_direction = "downstream"

    uuid = None
    username = None
    username_confirmed = False

    # used to stop people breaking the login process
    # by sending packets out-of-order or duplicated
    login_expecting = 0

    # the mojang 1.7.x client has a race condition where kicking immediately
    # after switching to "play" mode will cause a cast error in the client.
    # the fix is to set a deferred up which will fire when it's safe again
    safe_kick = None

    def __init__(self, factory, remote_addr):
        Protocol.__init__(self, factory, remote_addr)
        self.server_id    = crypto.make_server_id()
        self.verify_token = crypto.make_verify_token()


    ### Convenience functions -------------------------------------------------

    def switch_protocol_mode(self, mode):
        self.check_protocol_mode_switch(mode)

        if mode == "play":
            if self.factory.compression_threshold:
                # Send set compression
                self.send_packet("login_set_compression",
                                 self.buff_type.pack_varint(self.factory.compression_threshold))
                self.set_compression(self.factory.compression_threshold)

            # Send login success
            self.send_packet("login_success",
                self.buff_type.pack_string(self.uuid.to_hex()) +
                self.buff_type.pack_string(self.username))

            if self.protocol_version <= 5:
                def make_safe():
                    self.safe_kick.callback(None)
                    self.safe_kick = None

                def make_unsafe():
                    self.safe_kick = defer.Deferred()
                    self.tasks.add_delay(0.5, make_safe)

                make_unsafe()

        self.protocol_mode = mode

    def close(self, reason=None):
        """Closes the connection"""
        if not self.closed and reason is not None:
            # Kick the player if possible.
            if self.protocol_mode == "play":
                def real_kick(*a):
                    self.send_packet("disconnect",
                        self.buff_type.pack_chat(reason))
                    Protocol.close(self, reason)

                if self.safe_kick:
                    self.safe_kick.addCallback(real_kick)
                else:
                    real_kick()
            else:
                if self.protocol_mode == "login":
                    self.send_packet("login_disconnect",
                        self.buff_type.pack_chat(reason))
                Protocol.close(self, reason)
        else:
            Protocol.close(self, reason)


    ### Callbacks -------------------------------------------------------------

    def connection_lost(self, reason=None):
        """Called when the connection is lost"""
        if self.protocol_mode in ("login", "play"):
            self.factory.players.discard(self)
        Protocol.connection_lost(self, reason)

    def auth_ok(self, data):
        """Called when auth with mojang succeeded (online mode only)"""
        self.username_confirmed = True
        self.uuid = types.UUID.from_hex(data['id'])

        self.player_joined()

    def player_joined(self):
        """Called when the player joins the game"""
        Protocol.player_joined(self)

        self.logger.info("%s has joined." % self.username)

        self.switch_protocol_mode("play")

    def player_left(self):
        """Called when the player leaves the game"""
        Protocol.player_left(self)

        self.logger.info("%s has left." % self.username)

    ### Packet handlers -------------------------------------------------------

    def packet_handshake(self, buff):
        p_protocol_version = buff.unpack_varint()
        p_connect_host = buff.unpack_string()
        p_connect_port = buff.unpack("H")
        p_protocol_mode = buff.unpack_varint()

        mode = protocol_modes.get(p_protocol_mode, p_protocol_mode)
        self.switch_protocol_mode(mode)

        if mode == "login":
            if self.factory.force_protocol_version is not None:
                if p_protocol_version != self.factory.force_protocol_version:
                    self.close("Wrong protocol version")
            else:
                if p_protocol_version not in self.factory.minecraft_versions:
                    self.close("Unknown protocol version")

            if len(self.factory.players) >= self.factory.max_players:
                self.close("Server is full")
            else:
                self.factory.players.add(self)

        self.protocol_version = p_protocol_version
        self.connect_host = p_connect_host
        self.connect_port = p_connect_port

    def packet_login_start(self, buff):
        if self.login_expecting != 0:
            raise ProtocolError("Out-of-order login")

        self.username = buff.unpack_string()

        if self.factory.online_mode:
            self.login_expecting = 1

            # send encryption request

            # 1.7.x
            if self.protocol_version <= 5:
                pack_array = lambda a: self.buff_type.pack('h', len(a)) + a

            # 1.8.x
            else:
                pack_array = lambda a: self.buff_type.pack_varint(
                    len(a), max_bits=16) + a

            self.send_packet("login_encryption_request",
                self.buff_type.pack_string(self.server_id) +
                pack_array(self.factory.public_key) +
                pack_array(self.verify_token))

        else:
            self.login_expecting = None
            self.username_confirmed = True
            self.uuid = types.UUID.from_offline_player(self.username)

            self.player_joined()

    def packet_login_encryption_response(self, buff):
        if self.login_expecting != 1:
            raise ProtocolError("Out-of-order login")

        # 1.7.x
        if self.protocol_version <= 5:
            unpack_array = lambda b: b.read(b.unpack('h'))
        # 1.8.x
        else:
            unpack_array = lambda b: b.read(b.unpack_varint(max_bits=16))

        p_shared_secret = unpack_array(buff)
        p_verify_token = unpack_array(buff)

        shared_secret = crypto.decrypt_secret(
            self.factory.keypair,
            p_shared_secret)

        verify_token = crypto.decrypt_secret(
            self.factory.keypair,
            p_verify_token)

        self.login_expecting = None

        if verify_token != self.verify_token:
            raise ProtocolError("Verify token incorrect")

        # enable encryption
        self.cipher.enable(shared_secret)
        self.logger.debug("Encryption enabled")

        # make digest
        digest = crypto.make_digest(
            self.server_id.encode('ascii'),
            shared_secret,
            self.factory.public_key)

        # do auth
        deferred = auth.has_joined(
            self.factory.auth_timeout,
            digest,
            self.username)
        deferred.addCallbacks(self.auth_ok, self.auth_failed)

    def packet_status_request(self, buff):
        protocol_version = self.factory.force_protocol_version
        if protocol_version is None:
            protocol_version = self.protocol_version

        d = {
            "description": {
                "text":     self.factory.motd
            },
            "players": {
                "online":   len(self.factory.players),
                "max":      self.factory.max_players
            },
            "version": {
                "name":     self.factory.minecraft_versions.get(
                                protocol_version,
                                "???"),
                "protocol": protocol_version
            }
        }
        if self.factory.favicon is not None:
            with open(self.factory.favicon) as fd:
                d["favicon"] = "data:image/png;base64," + base64.encodestring(fd.read())

        # send status response
        self.send_packet("status_response", self.buff_type.pack_json(d))

    def packet_status_ping(self, buff):
        time = buff.unpack("Q")

        # send ping
        self.send_packet("status_pong", self.buff_type.pack("Q", time))
        self.close()


class ServerFactory(Factory):
    protocol = ServerProtocol
    force_protocol_version = None

    motd = "A Minecraft Server"
    max_players = 20
    favicon = None
    online_mode = True
    compression_threshold = 256

    def __init__(self):
        self.players = set()

        self.keypair = crypto.make_keypair()
        self.public_key = crypto.export_public_key(self.keypair)

    def listen(self, host, port=25565):
        reactor.listenTCP(port, self, interface=host)


import string
import sys
import logging
import zlib
from twisted.internet import protocol, reactor

from quarry.data import packets
from quarry.utils.crypto import Cipher
from quarry.utils.buffer import Buffer, BufferUnderrun
from quarry.utils.errors import ProtocolError
from quarry.utils.tasks import Tasks

PY3 = sys.version_info > (3,)

logging.basicConfig(format="%(name)s | %(levelname)s | %(message)s")

protocol_modes = {
    0: 'init',
    1: 'status',
    2: 'login',
    3: 'play'
}
protocol_modes_inv = dict(((v, k) for k, v in protocol_modes.items()))


class PacketDispatcher(object):
    def dispatch(self, lookup_args, buff):
        handler = getattr(self, "packet_%s" % "_".join(lookup_args), None)
        if handler is not None:
            handler(buff)
            return True
        return False

    def dump_packet(self, data):
        lines = ['Packet dump:']
        bytes_read = 0
        while len(data) > 0:
            data_line, data = data[:16], data[16:]

            l_hex = []
            l_str = []
            for i, c in enumerate(data_line):
                if PY3:
                    l_hex.append("%02x" % c)
                    c_str = data_line[i:i+1]
                    l_str.append(c_str if c_str in string.printable else ".")
                else:
                    l_hex.append("%02x" % ord(c))
                    l_str.append(c if c in string.printable else ".")

            l_hex.extend(['  '] * (16 - len(l_hex)))
            l_hex.insert(8, '')

            lines.append("%08x  %s  |%s|" % (
                bytes_read,
                " ".join(l_hex),
                "".join(l_str)))

            bytes_read += len(data_line)

        return "\n    ".join(lines + ["%08x" % bytes_read])


class Protocol(protocol.Protocol, PacketDispatcher, object):
    """Shared logic between the client and server"""

    #: Usually a reference to the :class:`Buffer` class. This is useful when
    #: constructing a packet payload for use in :meth:`send_packet`
    buff_type = None

    #: A reference to the logger
    logger = None

    #: A reference to a :class:`Tasks` instance. This object has methods for
    #: setting up repeating or delayed callbacks
    tasks = None

    recv_direction = None
    send_direction = None
    protocol_version = packets.default_protocol_version
    protocol_mode = "init"
    compression_threshold = None
    compression_enabled = False
    in_game = False
    closed = False

    def __init__(self, factory, remote_addr):
        self.factory = factory
        self.remote_addr = remote_addr

        self.buff_type = self.factory.buff_type
        self.recv_buff = self.buff_type()
        self.cipher = Cipher()
        self.tasks = Tasks()

        self.logger = logging.getLogger("%s{%s}" % (
            self.__class__.__name__,
            self.remote_addr.host))
        self.logger.setLevel(self.factory.log_level)

        self.connection_timer = self.tasks.add_delay(
            self.factory.connection_timeout,
            self.connection_timed_out)

        self.setup()

    ### Fix ugly twisted methods ----------------------------------------------

    def dataReceived(self, data):
        return self.data_received(data)

    def connectionMade(self):
        return self.connection_made()

    def connectionLost(self, reason=None):
        return self.connection_lost(reason)

    ### Convenience functions -------------------------------------------------

    def check_protocol_mode_switch(self, mode):
        transitions = [
            ("init", "status"),
            ("init", "login"),
            ("login", "play")
        ]

        if (self.protocol_mode, mode) not in transitions:
            raise ProtocolError("Cannot switch protocol mode from %s to %s"
                                % (self.protocol_mode, mode))

    def switch_protocol_mode(self, mode):
        self.check_protocol_mode_switch(mode)
        self.protocol_mode = mode

    def set_compression(self, compression_threshold):
        if not self.compression_enabled:
            self.compression_enabled = True
            self.logger.debug("Compression enabled")

        self.compression_threshold = compression_threshold
        self.logger.debug("Compression threshold set to %d bytes" % compression_threshold)

    def close(self, reason=None):
        """Closes the connection"""

        if not self.closed:
            if reason:
                reason = "Closing connection: %s" % reason
            else:
                reason = "Closing connection"

            if self.in_game:
                self.logger.info(reason)
            else:
                self.logger.debug(reason)

            self.transport.loseConnection()
            self.closed = True

    def log_packet(self, prefix, name):
        """Logs a packet at debug level"""

        self.logger.debug("Packet %s %s/%s" % (
            prefix,
            self.protocol_mode,
            name))

    ### General callbacks -----------------------------------------------------

    def setup(self):
        """Called when the Protocol's initialiser is finished"""

        pass

    def protocol_error(self, err):
        """Called when a protocol error occurs"""

        msg = "Protocol error: %s" % err
        self.logger.error(msg)
        self.close(msg)

    ### Connection callbacks --------------------------------------------------

    def connection_made(self):
        """Called when the connection is established"""

        self.logger.debug("Connection made")

    def connection_lost(self, reason=None):
        """Called when the connection is lost"""

        self.closed = True
        if self.in_game:
            self.player_left()
        self.logger.debug("Connection lost")

        self.tasks.stop_all()

    def connection_timed_out(self):
        """Called when the connection has been idle too long"""

        self.close("Connection timed out")

    ### Auth callbacks --------------------------------------------------------

    def auth_ok(self, data):
        """Called when auth with mojang succeeded (online mode only)"""

        pass

    def auth_failed(self, err):
        """Called when auth with mojang failed (online mode only)"""

        self.logger.warn("Auth failed: %s" % err.value)
        self.close("Auth failed: %s" % err.value)

    ### Player callbacks ------------------------------------------------------

    def player_joined(self):
        """Called when the player joins the game"""

        self.in_game = True

    def player_left(self):
        """Called when the player leaves the game"""

        pass

    ### Packet handling -------------------------------------------------------

    def data_received(self, data):
        # Decrypt data
        data = self.cipher.decrypt(data)

        # Add it to our buffer
        self.recv_buff.add(data)

        # Read some packets
        while not self.closed:
            # Save the buffer, in case we read an incomplete packet
            self.recv_buff.save()

            # Try to read a packet
            try:
                max_bits = 32 if self.protocol_mode == "play" else 21
                packet_length = self.recv_buff.unpack_varint(max_bits=max_bits)
                packet_body = self.recv_buff.read(packet_length)

            # Incomplete packet read, restore the buffer.
            except BufferUnderrun:
                self.recv_buff.restore()
                break

            # Load the packet body into a buffer
            packet_buff = self.buff_type()
            packet_buff.add(packet_body)

            try:  # Catch protocol errors
                try:  # Catch buffer overrun/underrun
                    if self.compression_enabled:
                        uncompressed_length = packet_buff.unpack_varint()

                        if uncompressed_length > 0:
                            data = zlib.decompress(packet_buff.read())
                            packet_buff = Buffer()
                            packet_buff.add(data)
                    ident = packet_buff.unpack_varint()
                    key = (
                        self.protocol_version,
                        self.protocol_mode,
                        self.recv_direction,
                        ident)
                    try:
                        name = packets.packet_names[key]
                    except KeyError:
                        raise ProtocolError("No name known for packet: %s"
                                            % (key,))
                    self.packet_received(packet_buff, name)

                except BufferUnderrun:
                    raise ProtocolError("Packet is too short!")

                if len(packet_buff) > 0:
                    raise ProtocolError("Packet is too long!")

            except ProtocolError as e:
                self.protocol_error(e)
                break

            # We've read a complete packet, so reset the inactivity timeout
            self.connection_timer.restart()

    def packet_received(self, buff, name):
        """
        Called when a packet is received from the remote. Usually this method
        dispatches the packet to a method named ``packet_<packet name>``, or
        calls :meth:`packet_unhandled` if no such methods exists. You might
        want to override this to implement your own dispatch logic or logging.
        """

        self.log_packet(". recv", name)

        dispatched = self.dispatch((name,), buff)

        if not dispatched:
            self.packet_unhandled(buff, name)

    def packet_unhandled(self, buff, name):
        """
        Called when a packet is received that is not hooked. The default
        implementation silently discards the packet.
        """

        buff.discard()

    def send_packet(self, name, data=b""):
        """Sends a packet to the remote."""

        if self.closed:
            return

        self.log_packet("# send", name)

        # Prepend ident
        key = (
            self.protocol_version,
            self.protocol_mode,
            self.send_direction,
            name)
        try:
            ident = packets.packet_idents[key]
        except KeyError:
            raise ProtocolError("No ID known for packet: %s" % (key,))
        data = Buffer.pack_varint(ident) + data

        if self.compression_enabled:
            # Compress data and prepend uncompressed data length
            if len(data) >= self.compression_threshold:
                data = Buffer.pack_varint(len(data)) + zlib.compress(data)
            else:
                data = Buffer.pack_varint(0) + data

        # Prepend packet length
        max_bits = 32 if self.protocol_mode == "play" else 21
        data = self.buff_type.pack_varint(len(data), max_bits=max_bits) + data

        # Encrypt
        data = self.cipher.encrypt(data)

        # Send
        self.transport.write(data)


class Factory(protocol.Factory, object):
    protocol = Protocol
    buff_type = Buffer
    log_level = logging.INFO
    connection_timeout = 30
    auth_timeout = 30

    minecraft_versions = packets.minecraft_versions

    def buildProtocol(self, addr):
        return self.protocol(self, addr)

    def run(self):
        reactor.run()

    def stop(self):
        reactor.stop()



import logging

from quarry.net.protocol import PacketDispatcher
from quarry.net.server import ServerFactory, ServerProtocol
from quarry.net.client import ClientFactory, ClientProtocol
from quarry.mojang.profile import Profile

#
# Rough diagram of the universe a quarry proxy usually operates in:
#               .
# outside world . server box
#               .
#  +--------+   .   +--------------------------------+       +--------+
#  | mojang | ----> |              QUARRY            | ----> | mojang |
#  | client | <---- | downstream | bridge | upstream | <---- | server |
#  +--------+   .   +--------------------------------+       +--------+
#               .          ^                   ^
#               .        server              client
#               .
#
# A quarry proxy has three main parts:
#  * The "downstream" (a server) which connects with the mojang client
#  * The "upstream" (a client) which connects with the mojang server
#  * The "bridge", which handles packet-passing between the above
#
# The downstream and upstream are broken down into two parts:
#  * DownstreamFactory - program-wide object that spawns a Downtime object for
#      each incoming connection
#  * Downstream - connection with the external client.
#  * UpstreamFactory - a simple class that spawns an Upstream
#  * Upstream - connection with the external server.
#
# When the user connects, the DownstreamFactory creates a Downstream object
#   to communicate with the external client. If we're running in online-mode,
#   we go through auth with mojang. Once the user is authed, we spawn
#   a Bridge object, which in turn spawns a UpstreamFactory. When the
#   connection is made external server, an Upstream object tells the Bridge
#   to enable "pass-through" mode. At this point, the Upstream and Downstream
#   are connected and exchange packets.
#
# You can inspect, drop, modify and forge packets in-transit by registering
#   handlers in the Bridge.
#

class Upstream(ClientProtocol):
    def packet_received_passthrough(self, buff, name):
        self.log_packet(". recv", name)
        self.factory.bridge.packet_received(
            buff,
            "downstream",
            name)

    def enable_passthrough(self):
        self.packet_received = self.packet_received_passthrough

    def setup(self):
        self.factory.bridge.upstream = self

    def player_joined(self):
        self.factory.bridge.upstream_connected()

    def connection_lost(self, reason=None):
        ClientProtocol.connection_lost(self, reason)
        self.factory.bridge.upstream_disconnected()


class UpstreamFactory(ClientFactory):
    protocol = Upstream
    bridge = None


class Bridge(PacketDispatcher):
    downstream_factory = None
    upstream_factory   = None
    downstream = None
    upstream   = None

    buff_type = None

    log_level = logging.INFO

    def __init__(self, downstream_factory, downstream):
        self.downstream_factory  = downstream_factory
        self.upstream_factory    = downstream_factory.upstream_factory_class()
        self.downstream = downstream
        self.upstream   = None

        self.buff_type = self.downstream_factory.buff_type

        self.logger = logging.getLogger("%s{%s}" % (
            self.__class__.__name__,
            self.downstream.username))
        self.logger.setLevel(self.log_level)

        # Set up client factory
        self.upstream_factory.bridge = self
        self.upstream_factory.buff_type = self.buff_type

        # If no upstream profile is set, generate an offline-mode profile
        if self.upstream_factory.profile is None:
            self.upstream_factory.profile = Profile()
            self.upstream_factory.profile.login_offline(
                self.downstream.username)

        # Connect to the server the client is requesting
        if self.downstream_factory.connect_host is None:
            self.connect_host = self.downstream.connect_host
            self.connect_port = self.downstream.connect_port
        else:
            self.connect_host = self.downstream_factory.connect_host
            self.connect_port = self.downstream_factory.connect_port

        self.setup()

    def setup(self):
        self.connect()

    def connect(self):
        self.upstream_factory.connect(
            self.connect_host,
            self.connect_port,
            "login",
            self.downstream.protocol_version)

    def upstream_connected(self):
        self.downstream.enable_passthrough()
        self.upstream.enable_passthrough()

    def upstream_disconnected(self):
        self.downstream.close("Lost connection to server.")

    def downstream_disconnected(self):
        if self.upstream:
            self.upstream.close()

    def packet_received(self, buff, direction, name):
        dispatched = self.dispatch((direction, name), buff)

        if not dispatched:
            self.packet_unhandled(buff, direction, name)

    def packet_unhandled(self, buff, direction, name):
        if direction == "downstream":
            self.downstream.send_packet(name, buff.read())
        elif direction == "upstream":
            self.upstream.send_packet(name, buff.read())

    def packet_downstream_set_compression(self, buff):
        self.upstream.set_compression(buff.unpack_varint())


class Downstream(ServerProtocol):
    bridge = None

    def packet_received_passthrough(self, buff, name):
        self.bridge.packet_received(
            buff,
            "upstream",
            name)

    def enable_passthrough(self):
        self.packet_received = self.packet_received_passthrough

    def player_joined(self):
        ServerProtocol.player_joined(self)
        self.bridge = self.factory.bridge_class(self.factory, self)

    def connection_lost(self, reason=None):
        ServerProtocol.connection_lost(self, reason)
        if self.bridge:
            self.bridge.downstream_disconnected()


class DownstreamFactory(ServerFactory):
    protocol = Downstream
    upstream_factory_class = UpstreamFactory
    connect_host = None
    connect_port = None
    bridge_class = Bridge



from quarry.utils import http


class AuthException(http.HTTPException):
    pass


def join(timeout, digest, access_token, uuid):
    return http.request(
        url=b"https://sessionserver.mojang.com/session/minecraft/join",
        timeout=timeout,
        err_type=AuthException,
        data={
            "accessToken": access_token,
            "selectedProfile": uuid.to_hex(with_dashes=False),
            "serverId": digest})

def has_joined(timeout, digest, username):
    return http.request(
        url=b"https://sessionserver.mojang.com/session/minecraft/hasJoined"
            b"?username=" + username.encode('ascii') + \
            b"&serverId=" + digest.encode('ascii'),
        timeout=timeout,
        err_type=AuthException,
        expect_content=True)


from twisted.internet import reactor
from twisted.internet.task import LoopingCall

class Tasks(object):
    def __init__(self):
        self._tasks = []

    def add_loop(self, time, callback, *args):
        task = LoopingCall(callback, *args)
        task.start(time, now=False)
        self._tasks.append(task)
        return task

    def add_delay(self, time, callback, *args):
        task = reactor.callLater(time, callback, *args)
        def stop():
            if task.active():
                task.cancel()
        def restart():
            if task.active():
                task.reset(time)
        task.restart = restart
        task.stop = stop
        self._tasks.append(task)
        return task

    def stop_all(self):
        while len(self._tasks) > 0:
            task = self._tasks.pop(0)
            task.stop()


import csv
import os

def _load():
    code_by_name = {}
    code_by_prop = {}
    csvpath = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        "..",
        "data",
        "chat_styles.csv"))
    with open(csvpath) as csvfile:
        reader = csv.reader(csvfile)
        for i, record in enumerate(reader):
            # Skip header
            if i == 0: continue

            code, name, prop = record
            code_by_name[name] = code
            if prop:
                code_by_prop[prop] = code

    return code_by_name, code_by_prop

code_by_name, code_by_prop = _load()


from twisted.internet import defer

from quarry.utils import http, types


class ProfileException(http.HTTPException):
    pass


class Profile(object):
    timeout = 30
    client_token = None
    access_token = None
    username = None
    uuid = None

    logged_in = False

    def _setData(self, data):
        self.client_token = data['clientToken']
        self.access_token = data['accessToken']
        self.username = data['selectedProfile']['name']
        self.uuid = types.UUID.from_hex(data['selectedProfile']['id'])

    def _req(self, endpoint, **data):
        return http.request(
            url=b"https://authserver.mojang.com/"+endpoint,
            timeout=self.timeout,
            err_type=ProfileException,
            data=data)

    def login_offline(self, username):
        self.username = username

    def login(self, username, password):
        d0 = defer.Deferred()

        def _callback(data):
            self._setData(data)
            self.logged_in = True
            d0.callback(data)

        def _errback(err):
            d0.errback(err)

        agent = {
            "name": "Minecraft",
            "version": 1
        }

        clientToken = "foo" #TODO

        d1 = self._req(b"authenticate",
            username = username,
            password = password,
            agent = agent,
            clientToken = clientToken
        )
        d1.addCallbacks(_callback, _errback)

        return d0

    def refresh(self, client_token, access_token, username, uuid):
        d0 = defer.Deferred()

        def _callback(data):
            self._setData(data)
            self.logged_in = True
            d0.callback(data)

        def _errback(err):
            d0.errback(err)

        d1 = self._req(b"refresh",
            clientToken = client_token,
            accessToken = access_token,
            selectedProfile = {
                "name": username,
                "id": uuid.to_hex()
            }
        )
        d1.addCallbacks(_callback, _errback)

        return d0

    def invalidate(self):
        return self._req(b"invalidate",
            clientToken = self.client_token,
            accessToken = self.access_token,
        )

    #TODO: validate
    #TODO: sign out


import csv
import os.path

def _load():
    default_protocol_version = 0
    minecraft_versions = {}
    packet_names = {}
    packet_idents = {}
    packet_ident = 0
    last_section = None
    csvpath = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        "..",
        "data",
        "packets.csv"))
    with open(csvpath) as csvfile:
        reader = csv.reader(csvfile)
        for i, record in enumerate(reader):
            # Skip header
            if i == 0: continue

            # Extract fields
            minecraft_version = record[0]
            protocol_version = int(record[1])
            protocol_mode = record[2]
            packet_direction = record[3]
            packet_name = record[4]

            # Check if ident should be reset
            section = (protocol_version, protocol_mode, packet_direction)
            if section != last_section:
                packet_ident = 0
            last_section = section

            # Update default protocol version
            default_protocol_version = max(default_protocol_version,
                                           protocol_version)

            # Update minecraft versions
            minecraft_versions[protocol_version] = minecraft_version

            # Update the packet dictionaries
            key = [protocol_version, protocol_mode, packet_direction]
            packet_names [tuple(key + [packet_ident])] = packet_name
            packet_idents[tuple(key + [packet_name ])] = packet_ident

            packet_ident += 1

    return default_protocol_version, minecraft_versions, \
           packet_names, packet_idents

default_protocol_version, minecraft_versions, \
packet_names, packet_idents = _load()



import sys
import csv
import re
import collections

PY3 = sys.version_info > (3,)
if PY3:
    import urllib.request
    urlread = lambda url: urllib.request.urlopen(url).read().decode('utf8')
else:
    import urllib2
    urlread = lambda url: urllib2.urlopen(url).read()

# Used to filter out pre-netty protocols
MINIMUM_OLDID = 5486

# Patterns
REGEX_WIKISOURCE = '<textarea.+?>(.+?)</textarea>'
REGEX_VERSIONS = "'''{{Minecraft Wiki\|([0-9\.]+)}}'''\n" \
                 '.+?\| (\d+)\n' \
                 '.+?\| \[http://wiki.vg/' \
                 '(?:index.php\?title=Protocol&amp;|Protocol\?)' \
                 'oldid=(\d+) page\]'
REGEX_SECTIONS = '\n(?P<level>=+)\s*(.+?)\s*(?P=level)'

def markdown_title(text, underline):
    return "%s\n%s\n\n" % (text, underline*len(text))

def download(page, oldid=None):
    url = "http://wiki.vg/index.php?title=%s&action=edit" % page
    if oldid is not None:
        url += "&oldid=%d" % oldid
    data = urlread(url)
    match = re.search(REGEX_WIKISOURCE, data, flags=re.DOTALL)
    return match.group(1)


def main():
    output_csv = csv.writer(open("packets.csv", "w"))
    output_csv.writerow(("minecraft_version", "protocol_version",
                         "protocol_mode", "packet_direction", "packet_ident",
                         "packet_name"))
    output_rst = open("../../docs/packet_names.rst", "w")
    output_rst_packets = collections.defaultdict(list)
    output_rst_version = None


    # Download the "Protocol version numbers" page
    p_versions = download("Protocol_version_numbers")

    for version in re.findall(REGEX_VERSIONS, p_versions):
        minecraft_version = version[0]
        protocol_version = int(version[1])
        oldid = int(version[2])

        # Ignore pre-netty protocols (quarry doesn't support them)
        if oldid >= MINIMUM_OLDID:

            # Grab packets from the "Protocol" page
            for packet in get_packets(oldid):
                protocol_mode, direction, ident, packet_name = packet
                # Write .csv
                output_csv.writerow([minecraft_version, protocol_version,
                                     protocol_mode, direction, ident,
                                     packet_name])

                # Collate for .rst
                if output_rst_version == None:
                    output_rst_version = minecraft_version
                if output_rst_version == minecraft_version:
                    output_rst_packets[packet_name].append(direction)

    # Write .rst
    output_rst.write(markdown_title("Packet Names", "="))
    output_rst.write("See the `Minecraft Coalition Wiki`_ for a details on "
                     "every packet.\n\n")
    output_rst.write(markdown_title("Minecraft %s" % output_rst_version, "-"))
    for full_name, modes in sorted(output_rst_packets.items()):
        output_rst.write("- ``%s`` (%s)\n" % (full_name, ", ".join(modes)))
    output_rst.write("\n")
    output_rst.write(".. _Minecraft Coalition Wiki: http://wiki.vg/Protocol")

def get_packets(oldid):
    # Nested wiki headings
    breadcrumbs = [None]*6

    # Download the "Protocol" page at the given revision
    p_protocol = download("Protocol", oldid)

    # Grab a list of sections, each with a heading and some content
    sections = re.split(REGEX_SECTIONS, p_protocol, flags=re.DOTALL)
    sections.pop(0)  # Discard content before first header

    while len(sections) > 0:
        heading_prefix = sections.pop(0)
        heading_name = sections.pop(0)
        content = sections.pop(0)

        # Update breadcrumbs
        breadcrumbs[len(heading_prefix)-2] = heading_name

        # Parse packets
        if heading_prefix == "====":
            # Extract packet ident
            match = re.search('\! Notes.+?0x([0-9a-zA-Z]{2})', content,
                              flags=re.DOTALL)
            ident = int(match.group(1), 16)

            # Extract protocol mode, direction and packet name
            protocol_mode = {
                "Handshaking": "init",
                "Status": "status",
                "Login": "login",
                "Play": "play"}[breadcrumbs[0]]
            direction = {
                "Serverbound": "upstream",
                "Clientbound": "downstream"}[breadcrumbs[1]]
            packet_name = re.sub('[^a-z]', '_', breadcrumbs[2].lower())

            # Ignore pre-netty ping packet
            if packet_name == "legacy_server_list_ping":
                continue

            # Clean up login packet names
            if packet_name.startswith("login_"):
                packet_name = packet_name[6:]

            # Prepend mode to name
            if protocol_mode in ("login", "status"):
                packet_name  = '%s_%s' % (protocol_mode, packet_name)

            yield protocol_mode, direction, ident, packet_name

if __name__ == '__main__':
    main()


from twisted.internet import reactor, protocol, defer

from quarry.net.protocol import Factory, Protocol, protocol_modes_inv
from quarry.mojang import auth
from quarry.utils import crypto
from quarry.utils.errors import ProtocolError


class ClientProtocol(Protocol):
    """This class represents a connection to a server"""

    recv_direction = "downstream"
    send_direction = "upstream"
    protocol_mode_next = None

    def __init__(self, factory, remote_addr):
        Protocol.__init__(self, factory, remote_addr)

    ### Convenience functions -------------------------------------------------

    def switch_protocol_mode(self, mode):
        self.check_protocol_mode_switch(mode)

        if mode in ("status", "login"):
            # Send handshake
            addr = self.transport.connector.getDestination()
            self.send_packet("handshake",
                self.buff_type.pack_varint(self.protocol_version) +
                self.buff_type.pack_string(addr.host) +
                self.buff_type.pack('H', addr.port) +
                self.buff_type.pack_varint(
                    protocol_modes_inv[self.protocol_mode_next]))

        self.protocol_mode = mode

        if mode == "status":
            # Send status request
            self.send_packet("status_request")

        elif mode == "login":
            # Send login start
            self.send_packet("login_start", self.buff_type.pack_string(
                self.factory.profile.username))


    ### Callbacks -------------------------------------------------------------

    def connection_made(self):
        """Called when the connection is established"""
        Protocol.connection_made(self)
        self.switch_protocol_mode(self.protocol_mode_next)

    def auth_ok(self, data):
        """
        Called if the Mojang session server responds to our query. Note that
        this method does not indicate that the server accepted our session; in
        this case :meth:`player_joined` is called.
        """

        # Send encryption response
        p_shared_secret = crypto.encrypt_secret(
            self.public_key,
            self.shared_secret)
        p_verify_token = crypto.encrypt_secret(
            self.public_key,
            self.verify_token)

        # 1.7.x
        if self.protocol_version <= 5:
            pack_array = lambda d: self.buff_type.pack('h', len(d)) + d

        # 1.8.x
        else:
            pack_array = lambda d: self.buff_type.pack_varint(
                len(d), max_bits=16) + d

        self.send_packet("login_encryption_response",
            pack_array(p_shared_secret) +
            pack_array(p_verify_token))

        # Enable encryption
        self.cipher.enable(self.shared_secret)
        self.logger.debug("Encryption enabled")

    def player_joined(self):
        """
        Called when we join the game. If the server is in online mode, this
        means the server accepted our session.
        """
        Protocol.player_joined(self)
        self.logger.info("Joined the game.")

    def player_left(self):
        """Called when we leave the game."""
        Protocol.player_left(self)
        self.logger.info("Left the game.")

    def status_response(self, data):
        """
        If we're connecting in "status" mode, this is called when the server
        sends us information about itself.
        """
        self.close()

    ### Packet handlers -------------------------------------------------------

    def packet_status_response(self, buff):
        p_data = buff.unpack_json()
        self.status_response(p_data)

    def packet_login_disconnect(self, buff):
        p_data = buff.unpack_chat()
        self.logger.warn("Kicked: %s" % p_data)
        self.close()

    def packet_login_encryption_request(self, buff):
        p_server_id    = buff.unpack_string()

        # 1.7.x
        if self.protocol_version <= 5:
            unpack_array = lambda b: b.read(b.unpack('h'))
        # 1.8.x
        else:
            unpack_array = lambda b: b.read(b.unpack_varint(max_bits=16))

        p_public_key   = unpack_array(buff)
        p_verify_token = unpack_array(buff)

        if not self.factory.profile.logged_in:
            raise ProtocolError("Can't log into online-mode server while using"
                                " offline profile")

        self.shared_secret = crypto.make_shared_secret()
        self.public_key = crypto.import_public_key(p_public_key)
        self.verify_token  = p_verify_token

        # make digest
        digest = crypto.make_digest(
            p_server_id.encode('ascii'),
            self.shared_secret,
            p_public_key)

        # do auth
        deferred = auth.join(
            self.factory.auth_timeout,
            digest,
            self.factory.profile.access_token,
            self.factory.profile.uuid)
        deferred.addCallbacks(self.auth_ok, self.auth_failed)

    def packet_login_success(self, buff):
        p_uuid = buff.unpack_string()
        p_username = buff.unpack_string()

        self.switch_protocol_mode("play")
        self.player_joined()

    def packet_login_set_compression(self, buff):
        self.set_compression(buff.unpack_varint())

    def packet_set_compression(self, buff):
        self.set_compression(buff.unpack_varint())

class ClientFactory(Factory, protocol.ClientFactory):
    protocol = ClientProtocol
    profile = None

    def connect(self, host, port=25565, protocol_mode_next="login",
                protocol_version=0):

        if protocol_mode_next == "status" or protocol_version > 0:
            self.protocol.protocol_mode_next = protocol_mode_next
            if protocol_version > 0:
                self.protocol.protocol_version = protocol_version
            reactor.connectTCP(host, port, self, self.connection_timeout)

        else:
            factory = ClientFactory()
            class PingProtocol(factory.protocol):
                def status_response(s, data):
                    s.close()
                    detected_version = int(data["version"]["protocol"])
                    if detected_version in self.minecraft_versions:
                        self.connect(host, port, protocol_mode_next,
                                     detected_version)
                    else:
                        pass #TODO

            factory.protocol = PingProtocol
            factory.connect(host, port, "status")
